#!perl
#
# Import Ogre3d Mesh XML file
# author: Timothy A. Graupmann
#
# This script will import an OGRE MESH XML file
#
# Good resources:
#
# C:\Program Files\Luxology\modo 201\Documentation\help\Scripting and Commands.pdf
#
#
#

#
# DEFAULTS
#
my $version = "1.0.0.0";
my $filter = "*.xml;*.xml";
my $ogreTools = "C:\\OGRETOOLS";

#
# VERSION OUTPUT
#
lxout ("Import Ogre3d XML Version: $version");

#
# AUTOFLUSH
#
$| = 1;

#
# EXAMINE ENVIRONMENT VARIABLES
#
#lxout "Available environment variables:";
#LuxPrint (join ("\n", %ENV));

#
# OVERRIDE FILTER DEFAULT WITH ENVIRONMENT VARIABLE
#
if (defined($ENV{"USERPROFILE"}))
{
  $filter = $ENV{"USERPROFILE"} . "\\Application Data\\Luxology\\Scripts\\*.mesh;*.xml";
}

#
# OPEN A MULTISELECTABLE FILE OPEN DIALOG
#
lx( "dialog.setup fileOpenMulti");
lx( "dialog.title {Import binary Ogre3d mesh file}");
lx( "dialog.filetype text");
lx( "dialog.result {$filter}");
lx( "dialog.open");

#
# EXTRACT THE FILES FROM THE DIALOG
#
my @files = lxq( "dialog.result ?" );

#
# ABORT IF NO FILES WERE SELECTED
#
if ($#files == -1)
{
  lxout "user aborted.";
}

#
# PROCESS EACH FILE
#
foreach my $filename (@files)
{
  lxout "Processing filename: $filename";
  
  if ($filename =~ m/.*[mM][eE][sS][hH]$/)
  {
    lxout "MESH FOUND, converting to XML...";

    #
    # OVERRIDE OGRETOOLS WITH ENVIRONMENT DEFAULTS
    #
    if (defined($ENV{"OGRETOOLS"}))
    {
      $ogreTools = $ENV{"OGRETOOLS"};
    }
    lxout "OGRETOOLS: $ogreTools";
    
    #
    # CHANGE TO THE FILENAME FOLDER
    #
    if ($filename =~ m/(.*)[\/\\]+[^\/^\\]*$/)    
    {
      my $filepath = $1;
      #lxout $filepath;      
      chdir $filepath;
      
      #
      # GENERATE NEW FILENAME
      #
      my $newFilename = $filename . "_" . scalar(time) . "mesh.xml";
      
      #
      # RUN THE CONVERTER
      #
      my $command = "\"" . $ogreTools . "\\OgreXMLConverter.exe\" \"$filename\" \"" . $newFilename . "\" |";
      lxout $command;
      open (FILEHANDLE, "| $command");
      while (<FILEHANDLE>)
      {
        # lxout "$_";
      }
      close (FILEHANDLE);
      
      #
      # PROCESS THE XML VERSION OF THE FILE
      #
      $filename = $newFilename;
    }
  }
  elsif ($filename =~ m/.*[xX][mM][lL]$/)
  {
    lxout "XML FOUND, GOOD";
  }
  else
  {
    lxout "UNKNOWN EXTENSION USED";
    lxout "user aborted.";
    return;
  }
  
  #
  # CACHE THE XML FILE
  #
  # my ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, $ctime, $blksize, $blocks) = stat($filename);
  # LuxPrint("$filename is $size bytes long.");
  open (FILEHANDLE, "<$filename") or
  	die ("Cannot open: $filename");

  my $g_content = "";
  seek(FILEHANDLE, 0, 0);
  while (<FILEHANDLE>)
  {
    # LuxPrint($_);
    $g_content .= $_;
  }
  close (FILEHANDLE);
  # LuxPrint("Read $size file bytes into content size: " . length($g_content));
  # LuxPrint($g_content);

  
  #
  # PARSE THE XML FILE
  #
  my $g_nestCount = 0;
  my %g_simpleXPath = ();
  my %g_uniqueXPaths = ();
  
  #init lux progress bar
  lxmonInit(length($g_content));
  
  #create new scene
  lx ("scene.new");
  my $vertex = 0;
  
  LuxPrint("PARSE EACH XML ELEMENT");
  while ($g_element = GetNextElement($g_content))
  {
    #somehow the modo scope is different so we need to remove the element from the content
    $g_content =~ s/$g_element//;
    
    #lux progress bar
    if (!lxmonStep(length($g_element)*2))
    {
      die( "User Abort" );
    }
    
    my $g_elementName = GetElementName($g_element);
  
    my $nestedWhitespace = GetNestedWhitespace($g_nestCount);
    #LuxPrint("$nestedWhitespace$g_element");
    #LuxPrint("$nestedWhitespace$g_elementName");
    #LuxPrint("$nestedWhitespace$g_elementName vs $g_element");
    
    CalculateXPath(\%g_simpleXPath, $g_nestCount, $g_element);
    #PrintHash(\%g_simpleXPath);
    
    my $xpath = GetXPath(\%g_simpleXPath, $g_nestCount);    
    #LuxPrint($xpath);
    
    BuildUniqueXPaths(\%g_uniqueXPaths, \%g_simpleXPath, $g_nestCount);
    #PrintHash(\%g_uniqueXPaths);
    
    $g_nestCount = CalculateNestCount($g_nestCount, $g_element);
    
    
    ###
    ### PARSE XML FOR LAYERS
    ###
    if ($xpath eq "mesh\\submeshes\\submesh")
    {
      #LuxPrint("Submesh detected.");
    }
    
    ###
    ### PARSE XML FOR COLORS
    ###
    if ($xpath eq "mesh\\submeshes\\submesh\\geometry\\vertexbuffer\\vertex\\colour_diffuse")
    {
      #LuxPrint("Vertex diffuse colors detected.");
    }
    
    ###
    ### PARSE XML FOR VERTECES
    ###
    if ($xpath eq "mesh\\submeshes\\submesh\\geometry\\vertexbuffer\\vertex\\position")
    {
      #LuxPrint("Vertex Position detected.");
      
      my $x = GetAttribute("x", $g_element);
      my $y = GetAttribute("y", $g_element);
      my $z = GetAttribute("z", $g_element);
      
      if ($vertex == 0)
      {
        LuxCommand("select.typeFrom vertex;edge;polygon;item;ptag [0]");
        LuxCommand("select.drop vertex");
        LuxCommand("select.invert");
        LuxCommand("hide.sel");
      }
      
      LuxCommand ("vert.new [$x][$y][$z]");
      
      if ($vertex == 2)
      {
        LuxCommand ("select.invert");
        LuxCommand ("poly.makeFace");
        $vertex = -1;
      }
      ++$vertex;
    }
    
    ###
    ### PARSE XML FOR TEXCOORDS
    ###
    if ($xpath eq "mesh\\submeshes\\submesh\\geometry\\vertexbuffer\\vertex\\texcoord")
    {
      #LuxPrint("Vertex Texcoords detected.");
    }
    
    ###
    ### BUILD FACE
    ###
    if ($g_element eq "</submesh>")
    {
      LuxCommand ("unhide");
    }
  }
  LuxPrint("Unique XPaths:");
  PrintHash(\%g_uniqueXPaths);
}


# return;


######
###### HELPERS
######

# lxout can only print so much per line
sub LuxPrint
{
  my $luxPrint = shift @_;
  my @luxPrint = split(/\n/, $luxPrint);
  foreach my $line (@luxPrint)
  {
    lxout $line;
    #print "$line\n";
  }
}

sub LuxCommand
{
  my $luxCommand = shift @_;
  lx ($luxCommand);
  LuxPrint ($luxCommand);
}

sub PrintHash
{
  my $hash_ref = shift @_;
  foreach my $key (sort %$hash_ref)
  {
    if ($key ne "")
    {
      LuxPrint ($key);
    }
  }
}

######
###### SIMPLE XML PARSING ROUTINES, HAHA HACK HACK HACK
######

sub GetNextElement
{
  my $getNextElement = shift @_;  
  if ($getNextElement =~ m/[<]{1}([^>]+)[>]{1}/g)
  {
    return "<$1>";
  }
  # LuxPrint("No more elements.");
  return "";
}

#getAttribute, #element
sub GetAttribute
{
  my $getAttribute = shift @_;
  my $element = shift @_;
  
  #print "GetAttribute: [$getAttribute]\n";

  my @blob = split(/ /, $element);
  
  foreach my $innerBlob (@blob)
  {
    # LuxPrint("$innerBlob");
    my @nvp = split(/="/, $innerBlob);
    if (scalar(@nvp) > 1)
    {
      if ($nvp[0] eq $getAttribute)
      {
        my @valueEnd = split(/"/, $nvp[1]);
        if (scalar(@valueEnd) > 0)
        {
          return $valueEnd[0];
        }
      }
    }
  }
  
}

# $nestCount, $element
sub CalculateNestCount
{
  my $nestCount = shift @_;
  #LuxPrint("nestCount: $nestCount");
  
  my $element = shift @_;
  #LuxPrint("element: $element");
  
  if (index($element, "/>") != -1)
  {
  }
  elsif (index($element, "</") != -1)
  {
    $nestCount--;
  }
  else
  {
    $nestCount++;
  }
  return $nestCount;
}

# $nestCount
sub GetNestedWhitespace
{
  my $nestCount = shift @_;  
  #LuxPrint("nestCount: $nestCount");
  
  my $nestedWhitespace = "";
  for(my $index = 0; $index < $nestCount; ++$index)
  {
    $nestedWhitespace .= " ";
  }
  return $nestedWhitespace;
}

#element
sub GetElementName
{
  my $element = shift @_;

  my @splitElement = split("", $element);
  my $offset = 0;
  for(my $index = 0; $index < length($element); ++$index)
  {
    if ($splitElement[$index] eq "<")
    {
      ++$offset;
    }
    elsif ($splitElement[$index] eq "/")
    {
      if ($offset == 1)
      {
       ++$offset;
      }
      else
      {
        return substr($element, $offset, $index-$offset);
      }
    }
    elsif ($splitElement[$index] eq " ")
    {
      return substr($element, $offset, $index-$offset);
    }
    elsif ($splitElement[$index] eq ">")
    {
      return substr($element, $offset, $index-$offset);
    }
  }
  return $element;
}

# %simpleXPath, $nestCount, $element
sub CalculateXPath
{
  my $simpleXPath_ref = shift @_;
  #LuxPrint("simpleXPath: $simpleXPath_ref");
  #PrintHash($simpleXPath_ref);
  
  my $nestCount = shift @_;
  #LuxPrint("nestCount: $nestCount");
  
  my $element = shift @_;
  #LuxPrint("element: $element");
  
  my $elementName = GetElementName($element);
  $simpleXPath_ref->{$nestCount} = $elementName;
  #LuxPrint("Before Ret: " . $simpleXPath_ref->{$nestCount});
}

# %simpleXPath, $nestCount
sub GetXPath
{
  my $simpleXPath_ref = shift @_;
  #LuxPrint("simpleXPath_ref: $simpleXPath_ref");
  
  my $nestCount = shift @_;
  #LuxPrint("nestCount: $nestCount");
  
  my $ret = "";
  for(my $index = 0; $index <= $nestCount; ++$index)
  {
    if ($index > 0)
    {
      $ret .= "\\";
    }
    $ret .= $simpleXPath_ref->{$index};
  }
  return $ret;
}

# %uniqueXPaths, %simpleXPath, $nestCount
sub BuildUniqueXPaths
{
  my $uniqueXPaths_ref = shift @_;
  #LuxPrint("uniqueXPaths_ref: $uniqueXPaths_ref");
  
  my $simpleXPath_ref = shift @_;
  #LuxPrint("simpleXPath_ref: $simpleXPath_ref");
  
  my $nestCount = shift @_;
  #LuxPrint("nestCount: $nestCount");
  
  my $xpath = GetXPath($simpleXPath_ref, $nestCount);
  if (!defined($uniqueXPaths_ref->{$xpath}))
  {
    $uniqueXPaths_ref->{$xpath} = "";
    #LuxPrint("Added [$xpath]\n");
  }
}